//
// module data
// (c) 10.11.2007 Ingo D. Rullhusen
//

#include <stdio.h>
#include <math.h>
#include <float.h>

#include <vector>
#include <algorithm>


#include <iostream>


#include "msh.h"

#include "module_data.h"
#include "rotmat.h"


using namespace std;


static const double DIST_EPSILON = 1e-10;


//
// TETRAEDER
//
TETRAEDER::TETRAEDER()
: p1(-1), p2(-1), p3(-1), p4(-1),
  Tb1(-1), Tb2(-1), Tb3(-1), Tb4(-1)
{ sign.sp1 = sign.sp2 = sign.sp3 = sign.sp4 = 0; }

//
// Tbfkt
//
Tbfkt::Tbfkt()
: tetra1(-1), tetra2(-1),
  trip1(-1), trip2(-1), trip3(-1),
  sp1(-1), sp2(-1)
{}

double Tbfkt::triangle_area( POINT_LIST plist ) const
{
   double a[3], b[3], nx, ny, nz;

   a[0] = plist[ trip2 ].x - plist[ trip1 ].x;
   a[1] = plist[ trip2 ].y - plist[ trip1 ].y;
   a[2] = plist[ trip2 ].z - plist[ trip1 ].z;

   b[0] = plist[ trip3 ].x - plist[ trip1 ].x;
   b[1] = plist[ trip3 ].y - plist[ trip1 ].y;
   b[2] = plist[ trip3 ].z - plist[ trip1 ].z;

   nx = a[1]*b[2] - a[2]*b[1];
   ny = a[2]*b[0] - a[0]*b[2];
   nz = a[0]*b[1] - a[1]*b[0];

   return( 0.5*sqrt(nx*nx + ny*ny + nz*nz) );
}



//
// ModuleData
//

ModuleData::ModuleData()
{
   unit = 1.0;
   cpx = cpy = cpz = 0.0;
   theta = phi = psi = 0.0;

   msh_file_name = "untitled.msh";

   no_material = 0;
   ml = NULL;

   plist.clear();

   tetralist.clear();

   bfktlist.clear();


   grf = NULL;


   probe_plist.clear();
}

ModuleData::ModuleData( const ModuleData *d )
{

   unit = d->unit;
   cpx = d->cpx;
   cpy = d->cpy;
   cpz = d->cpz;

   theta = d->theta;
   phi = d->phi;
   psi = d->psi;

   msh_file_name = d->msh_file_name;

   no_material = d->no_material;
   ml = new MATERIAL[ no_material ];
   for ( int i=0; i<no_material; i++ ) ml[i] = d->ml[i];


   plist = d->plist;

   tetralist = d->tetralist;

   bfktlist = d->bfktlist;


   if ( d->grf )
      grf = new GREEN( *d->grf );
   else
      grf = NULL;


   probe_plist = d->probe_plist;
}


ModuleData::~ModuleData()
{
   delete [] ml;

   plist.clear();

   tetralist.clear();

   bfktlist.clear();

   delete grf;

   probe_plist.clear();
}



class SORTpoint
{
   public:
      int no;
      double x, y, z;

      SORTpoint( int no_in=-1, double x_in=0.0, double y_in=0.0, double z_in=0.0 )
      : no(no_in), x(x_in), y(y_in), z(z_in)
      {}

      bool operator> ( const SORTpoint& sp2 ) const
      {
         if ( x - sp2.x > DBL_EPSILON )
            return(true);
         else if ( fabs(x - sp2.x) < DBL_EPSILON )
            if ( y - sp2.y > DBL_EPSILON )   
               return(true);
            else if ( fabs(y - sp2.y) < DBL_EPSILON )
               if ( z > sp2.z )   
                  return(true);

         return(false);
      };

      bool operator< ( const SORTpoint& sp2 ) const
      {
         if ( x - sp2.x < -DBL_EPSILON )
            return(true);
         else if ( fabs(x - sp2.x) < DBL_EPSILON )
            if ( y - sp2.y < -DBL_EPSILON )
               return(true);
            else if ( fabs(y - sp2.y) < DBL_EPSILON )
               if ( z < sp2.z )
                  return(true);

         return(false);
      };

      operator POINT () const { POINT p; p.x=x; p.y=y; p.z=z; return(p); }
};

class SORTdata
{
   public:
      int p1, p2, p3;
      int p4, t;

      SORTdata( int p1_in=-1, int p2_in=-1, int p3_in=-1, int p4_in=-1, int t_in =-1 )
      : p1(p1_in), p2(p2_in), p3(p3_in), p4(p4_in), t(t_in)
      {}

      bool operator> ( const SORTdata& sd2 ) const
      {
         if ( p1 > sd2.p1 )
            return(true);
         else if ( p1 == sd2.p1 )
            if ( p2 > sd2.p2 )
               return(true);
            else if ( p2 == sd2.p2 )
               if ( p3 > sd2.p3 )
                  return(true);

         return(false);
      };

      bool operator< ( const SORTdata& sd2 ) const
      {
         if ( p1 < sd2.p1 )
            return(true);
         else if ( p1 == sd2.p1 )
            if ( p2 < sd2.p2 )
               return(true);
            else if ( p2 == sd2.p2 )
               if ( p3 < sd2.p3 )
                  return(true);

         return(false);
      };
};


typedef struct { int p1, p2, p3, p4, mat; } tetra_temp;



void ModuleData::setup_geo()
{
   int m, n;
   double x, y, z, rmat[3][3];

   rotmat( rmat, unit, psi, 0.0, 0.0, theta, phi );

//
// read foreign
//
   FILE *mshfile;

   double version;
   map<int, node*> nodes;
   vector<element*> elements;

   mshfile = fopen( msh_file_name.c_str(), "r" );
   if ( mshfile == NULL )
   {
      throw( "cannot open msh file" );
   }

   version = readMesh( mshfile, nodes, elements );
   mshfile = NULL;  // readMesh close

// transfer to internal
   vector<SORTpoint> pl_t;

   pl_t.resize( nodes.size() );

   n = 0;
   for ( map<int, node*>::iterator it=nodes.begin();
         it != nodes.end();
         ++it )
   {
      it->second->_num = n;

      pl_t[n].no = n;
      x = it->second->_x * unit;
      y = it->second->_y * unit;
      z = it->second->_z * unit;

      pl_t[n].x = rmat[0][0]*x + rmat[0][1]*y + rmat[0][2]*z + cpx;
      pl_t[n].y = rmat[1][0]*x + rmat[1][1]*y + rmat[1][2]*z + cpy;
      pl_t[n].z = rmat[2][0]*x + rmat[2][1]*y + rmat[2][2]*z + cpz;

      ++n;
   }

   tetra_temp tlt;
   vector<tetra_temp> tl_t;

   tl_t.clear();
   for ( vector<element*>::iterator it=elements.begin();
         it != elements.end();
         ++it )
   {
      if ( (*it)->_type != 4 ) continue;

      for ( m=0; m<no_material; m++ )
      {
         if ( ml[m].phys_entity == (*it)->_physical )
         {
            tlt.p1 = (*it)->_nodes[0]->_num;
            tlt.p2 = (*it)->_nodes[1]->_num;
            tlt.p3 = (*it)->_nodes[2]->_num;
            tlt.p4 = (*it)->_nodes[3]->_num;

            tlt.mat = m;

            tl_t.push_back(tlt);
         }
      }
   }

   for ( map<int, node*>::iterator it=nodes.begin();
         it != nodes.end();
         ++it )
   {
      delete it->second;
      it->second = NULL;
   }
   nodes.clear();

   for ( vector<element*>::iterator it=elements.begin();
         it != elements.end();
         ++it )
   {
      delete *it;
      *it = NULL;
   }
   elements.clear();



//
// setup geometry
//

// remove multiple points (pl)
   sort( pl_t.begin(), pl_t.end() );

   int *permute, p;
   permute = new int[ pl_t.size() ];

   vector<SORTpoint>::iterator isp1, isp2;
   isp1 = pl_t.begin();

   p = 0;
   permute[isp1->no] = p;

   for ( isp2 = isp1 + 1;
         isp2 != pl_t.end();
         isp1++, isp2++ )   
   {
      if ( HYPOT( isp1->x-isp2->x, isp1->y-isp2->y, isp1->z-isp2->z ) > DIST_EPSILON )
      {
         p++;
         permute[isp2->no] = p;
      }   
      else
      {   
         permute[isp2->no] = p;
      }
   }

   plist.resize( p+1 );
   for ( unsigned int i=0; i<pl_t.size(); i++ )
   {
      if ( permute[ pl_t[i].no ] < 0 || permute[ pl_t[i].no ] > p ) throw( "point index problem" );
      plist[ permute[ pl_t[i].no ] ] = pl_t[i];
   }

   pl_t.clear();


   for ( vector<tetra_temp>::iterator it = tl_t.begin();
         it != tl_t.end();
         it++ )
   {
      it->p1 = permute[it->p1];
      it->p2 = permute[it->p2];
      it->p3 = permute[it->p3];
      it->p4 = permute[it->p4];

   }

   delete [] permute;

//
// search for basis functions (Tbfktl/tetral)
   tetralist.resize( tl_t.size() );

   {
   TETRAEDER_LIST::iterator iT;
   vector<tetra_temp>::iterator it;

   for ( iT = tetralist.begin(), it = tl_t.begin();
         iT != tetralist.end();
         ++iT, ++it )
   {
      iT->p1 = it->p1;
      iT->p2 = it->p2;
      iT->p3 = it->p3;
      iT->p4 = it->p4;

      iT->material = it->mat;
   }

   }

   tl_t.clear();

   if ( tetralist.size() < 1 ) throw( "no tetrahedral region found" );

   int t;
   vector<int> tpl_t(3);
   vector<SORTdata> SDlist;

   SDlist.clear();
   for ( TETRAEDER_LIST::iterator iT = tetralist.begin();
         iT != tetralist.end();
         ++iT )
   {
      t = iT - tetralist.begin();

// triangle 1
      tpl_t[0] = iT->p2;
      tpl_t[1] = iT->p3;
      tpl_t[2] = iT->p4;
      sort( tpl_t.begin(), tpl_t.end() );

      SDlist.push_back( SORTdata( tpl_t[0], tpl_t[1], tpl_t[2],  iT->p1, t ) );

// triangle 2
      tpl_t[0] = iT->p1;
      tpl_t[1] = iT->p3;
      tpl_t[2] = iT->p4;
      sort( tpl_t.begin(), tpl_t.end() );

      SDlist.push_back( SORTdata( tpl_t[0], tpl_t[1], tpl_t[2],  iT->p2, t ) );

// triangle 3
      tpl_t[0] = iT->p1;
      tpl_t[1] = iT->p2;
      tpl_t[2] = iT->p4;
      sort( tpl_t.begin(), tpl_t.end() );

      SDlist.push_back( SORTdata( tpl_t[0], tpl_t[1], tpl_t[2],  iT->p3, t ) );

// triangle 4
      tpl_t[0] = iT->p1;
      tpl_t[1] = iT->p2;
      tpl_t[2] = iT->p3;
      sort( tpl_t.begin(), tpl_t.end() );

      SDlist.push_back( SORTdata( tpl_t[0], tpl_t[1], tpl_t[2],  iT->p4, t ) );
   }

   sort( SDlist.begin(), SDlist.end() );

   bool cont;
   int bfktno;
   Tbfkt bfkt_t;
   vector<SORTdata>::iterator sd1, sd2;

   bfktno = 0;
   bfktlist.clear();

   cont = false;
   sd1 = SDlist.begin();
   while( sd1->p1<0 || sd1->p2<0 || sd1->p3<0 || sd1->p4<0 || sd1->t<0 ) ++sd1;
   for ( sd2 = sd1 + 1;
         sd2 != SDlist.end();
         ++sd1, ++sd2 )
   {
      if ( cont )
      {
         cont = false;
      }
      else if ( sd1->p1 == sd2->p1
                && sd1->p2 == sd2->p2
                   && sd1->p3 == sd2->p3 )
      {
         cont = true;

         bfkt_t.tetra1 = sd1->t;
         bfkt_t.tetra2 = sd2->t;
         bfkt_t.trip1 = sd1->p1;
         bfkt_t.trip2 = sd1->p2;
         bfkt_t.trip3 = sd1->p3;
         bfkt_t.sp1 = sd1->p4;
         bfkt_t.sp2 = sd2->p4;
         bfktlist.push_back( bfkt_t );

// tetraeder 1
         if ( tetralist[sd1->t].p1 == sd1->p4 )
         {
            tetralist[sd1->t].Tb1 = bfktno;
            tetralist[sd1->t].sign.sp1 = +1;
         }
         else if ( tetralist[sd1->t].p2 == sd1->p4 )
         {
            tetralist[sd1->t].Tb2 = bfktno;
            tetralist[sd1->t].sign.sp2 = +1;
         }
         else if ( tetralist[sd1->t].p3 == sd1->p4 )
         {
            tetralist[sd1->t].Tb3 = bfktno;
            tetralist[sd1->t].sign.sp3 = +1;
         }
         else if ( tetralist[sd1->t].p4 == sd1->p4 )
         {
            tetralist[sd1->t].Tb4 = bfktno;
            tetralist[sd1->t].sign.sp4 = +1;
         }
         else
         {
            cerr << "Error in DielectricModule ModuleData::ModuleData(): internal line "
                 << __LINE__ << "!" << endl;
            exit(1);
         }

// tetraeder 2
         if ( tetralist[sd2->t].p1 == sd2->p4 )
         {
            tetralist[sd2->t].Tb1 = bfktno;
            tetralist[sd2->t].sign.sp1 = -1;
         }
         else if ( tetralist[sd2->t].p2 == sd2->p4 )
         {
            tetralist[sd2->t].Tb2 = bfktno;
            tetralist[sd2->t].sign.sp2 = -1;
         }
         else if ( tetralist[sd2->t].p3 == sd2->p4 )
         {
            tetralist[sd2->t].Tb3 = bfktno;
            tetralist[sd2->t].sign.sp3 = -1;
         }
         else if ( tetralist[sd2->t].p4 == sd2->p4 )
         {
            tetralist[sd2->t].Tb4 = bfktno;
            tetralist[sd2->t].sign.sp4 = -1;
         }
         else
         {
            cerr << "Error in DielectricModule ModuleData::ModuleData(): internal line "
                 << __LINE__ << "!" << endl;
            exit(1);
         }

         ++bfktno;
      }
      else
      {
         bfkt_t.tetra1 = sd1->t;
         bfkt_t.tetra2 = -1;
         bfkt_t.trip1 = sd1->p1;
         bfkt_t.trip2 = sd1->p2;
         bfkt_t.trip3 = sd1->p3;
         bfkt_t.sp1 = sd1->p4;
         bfkt_t.sp2 = -1;
         bfktlist.push_back( bfkt_t );

// tetraeder 1
         if ( tetralist[sd1->t].p1 == sd1->p4 )
         {
            tetralist[sd1->t].Tb1 = bfktno;
            tetralist[sd1->t].sign.sp1 = +1;
         }
         else if ( tetralist[sd1->t].p2 == sd1->p4 )
         {
            tetralist[sd1->t].Tb2 = bfktno;
            tetralist[sd1->t].sign.sp2 = +1;
         }
         else if ( tetralist[sd1->t].p3 == sd1->p4 )
         {
            tetralist[sd1->t].Tb3 = bfktno;
            tetralist[sd1->t].sign.sp3 = +1;
         }
         else if ( tetralist[sd1->t].p4 == sd1->p4 )
         {
            tetralist[sd1->t].Tb4 = bfktno;
            tetralist[sd1->t].sign.sp4 = +1;
         }
         else
         {
            cerr << "Error in DielectricModule ModuleData::ModuleData(): internal line "
                 << __LINE__ << "!" << endl;
            exit(1);
         }

// tetraeder 2
// not present

         ++bfktno;
      }
   }   // end: for ( ... ++sd1, ++sd2 )

   if ( sd1 != SDlist.end() )
   {
      bfkt_t.tetra1 = sd1->t;
      bfkt_t.tetra2 = -1;
      bfkt_t.trip1 = sd1->p1;
      bfkt_t.trip2 = sd1->p2;
      bfkt_t.trip3 = sd1->p3;
      bfkt_t.sp1 = sd1->p4;
      bfkt_t.sp2 = -1;
      bfktlist.push_back( bfkt_t );

// tetraeder 1
      if ( tetralist[sd1->t].p1 == sd1->p4 )
      {
         tetralist[sd1->t].Tb1 = bfktno;
         tetralist[sd1->t].sign.sp1 = +1;
      }
      else if ( tetralist[sd1->t].p2 == sd1->p4 )
      {
         tetralist[sd1->t].Tb2 = bfktno;
         tetralist[sd1->t].sign.sp2 = +1;
      }
      else if ( tetralist[sd1->t].p3 == sd1->p4 )
      {
         tetralist[sd1->t].Tb3 = bfktno;
         tetralist[sd1->t].sign.sp3 = +1;
      }
      else if ( tetralist[sd1->t].p4 == sd1->p4 )
      {
         tetralist[sd1->t].Tb4 = bfktno;
         tetralist[sd1->t].sign.sp4 = +1;
      }
      else
      {
         cerr << "Error in DielectricModule ModuleData::ModuleData(): internal line "
              << __LINE__ << "!" << endl;
         exit(1);
      }

// tetraeder 2
// not present

      ++bfktno;
   }

   SDlist.clear();

}


#ifdef HAVE_LIBGL
void ModuleData::ogl_triangles( GLuint list )
{
   const GLfloat clr_1[4] = { 0.12, 0.90, 0.90, 0.5 };
   const GLfloat clr_2[4] = { 0.12, 0.90, 0.90, 0.1 };
   const GLfloat clr_3[4] = { 0.90, 0.05, 0.05, 0.5 };

   double u[3], v[3], w[3], a;

//   glDisable( GL_DEPTH_TEST );
   glEnable( GL_BLEND );
   glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );


//
// inner borders
   glMaterialfv( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, clr_3 );

   glBegin(GL_TRIANGLES);

   for ( Tbfkt_LIST::iterator it = bfktlist.begin();
         it != bfktlist.end();  
         it++ )
   {
      if ( it->tetra2 < 0 ) continue;

      if ( tetralist[ it->tetra1 ].material != tetralist[ it->tetra2 ].material )
      {
         u[0] = plist[ it->trip1 ].x -  plist[ it->trip2 ].x;
         u[1] = plist[ it->trip1 ].y -  plist[ it->trip2 ].y;
         u[2] = plist[ it->trip1 ].z -  plist[ it->trip2 ].z;

         v[0] = plist[ it->trip3 ].x -  plist[ it->trip2 ].x;
         v[1] = plist[ it->trip3 ].y -  plist[ it->trip2 ].y;
         v[2] = plist[ it->trip3 ].z -  plist[ it->trip2 ].z;

         w[0] = u[1]*v[2] - u[2]*v[1];
         w[1] = u[2]*v[0] - u[0]*v[2];
         w[2] = u[0]*v[1] - u[1]*v[0];

         u[0] = plist[ it->sp2 ].x -  plist[ it->sp1 ].x;
         u[1] = plist[ it->sp2 ].y -  plist[ it->sp1 ].y;
         u[2] = plist[ it->sp2 ].z -  plist[ it->sp1 ].z;

         if ( w[0]*u[0]+w[1]*u[1]+w[2]*u[2] > 0 )
         {
            a = sqrt( w[0]*w[0] + w[1]*w[1] + w[2]*w[2] );
            glNormal3f( w[0]/a, w[1]/a, w[2]/a );

            glVertex3f( plist[it->trip2].x, plist[it->trip2].y, plist[it->trip2].z );
            glVertex3f( plist[it->trip1].x, plist[it->trip1].y, plist[it->trip1].z );
            glVertex3f( plist[it->trip3].x, plist[it->trip3].y, plist[it->trip3].z );
         }
         else
         {
            a = -sqrt( w[0]*w[0] + w[1]*w[1] + w[2]*w[2] );
            glNormal3f( w[0]/a, w[1]/a, w[2]/a );

            glVertex3f( plist[it->trip1].x,  plist[it->trip1].y, plist[it->trip1].z );
            glVertex3f( plist[it->trip2].x,  plist[it->trip2].y, plist[it->trip2].z );
            glVertex3f( plist[it->trip3].x,  plist[it->trip3].y, plist[it->trip3].z );
         }
      }
   }

   glEnd();

//
// outer border
   glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, clr_1 );
   glMaterialfv( GL_BACK, GL_AMBIENT_AND_DIFFUSE, clr_2 );

   glBegin(GL_TRIANGLES);

   for ( Tbfkt_LIST::iterator it = bfktlist.begin();
         it != bfktlist.end();  
         it++ )
   {
      if ( it->tetra2 < 0 )
      {
         u[0] = plist[ it->trip1 ].x -  plist[ it->trip2 ].x;
         u[1] = plist[ it->trip1 ].y -  plist[ it->trip2 ].y;
         u[2] = plist[ it->trip1 ].z -  plist[ it->trip2 ].z;

         v[0] = plist[ it->trip3 ].x -  plist[ it->trip2 ].x;
         v[1] = plist[ it->trip3 ].y -  plist[ it->trip2 ].y;
         v[2] = plist[ it->trip3 ].z -  plist[ it->trip2 ].z;

         w[0] = u[1]*v[2] - u[2]*v[1];
         w[1] = u[2]*v[0] - u[0]*v[2];
         w[2] = u[0]*v[1] - u[1]*v[0];

         u[0] = plist[ it->trip2 ].x -  plist[ it->sp1 ].x;
         u[1] = plist[ it->trip2 ].y -  plist[ it->sp1 ].y;
         u[2] = plist[ it->trip2 ].z -  plist[ it->sp1 ].z;

         if ( w[0]*u[0]+w[1]*u[1]+w[2]*u[2] > 0 )
         {
            a = sqrt( w[0]*w[0] + w[1]*w[1] + w[2]*w[2] );
            glNormal3f( w[0]/a, w[1]/a, w[2]/a );

            glVertex3f( plist[it->trip2].x, plist[it->trip2].y, plist[it->trip2].z );
            glVertex3f( plist[it->trip1].x, plist[it->trip1].y, plist[it->trip1].z );
            glVertex3f( plist[it->trip3].x, plist[it->trip3].y, plist[it->trip3].z );
         }
         else
         {
            a = -sqrt( w[0]*w[0] + w[1]*w[1] + w[2]*w[2] );
            glNormal3f( w[0]/a, w[1]/a, w[2]/a );

            glVertex3f( plist[it->trip1].x,  plist[it->trip1].y, plist[it->trip1].z );
            glVertex3f( plist[it->trip2].x,  plist[it->trip2].y, plist[it->trip2].z );
            glVertex3f( plist[it->trip3].x,  plist[it->trip3].y, plist[it->trip3].z );
         }
      }
   }

   glEnd();


   glDisable( GL_BLEND );
//   glEnable( GL_DEPTH_TEST );
}
#endif


void ModuleData::tetraeder_center( int tno, POINT &cp ) const
{
#ifdef BOUNDARY_CHECK
   if ( tno < 0 || tno >= (int)tetralist.size() )
   {
      cerr << "Error in ModuleData::tetraeder_center(): tno out of bounds!"
           << endl;
      exit(1);
   }
#endif
   cp.x = ( plist[ tetralist[tno].p1 ].x
            + plist[ tetralist[tno].p2 ].x
              + plist[ tetralist[tno].p3 ].x
                + plist[ tetralist[tno].p4 ].x ) * 0.25;
   cp.y = ( plist[ tetralist[tno].p1 ].y
            + plist[ tetralist[tno].p2 ].y
              + plist[ tetralist[tno].p3 ].y
                + plist[ tetralist[tno].p4 ].y ) * 0.25;
   cp.z = ( plist[ tetralist[tno].p1 ].z
            + plist[ tetralist[tno].p2 ].z
              + plist[ tetralist[tno].p3 ].z
                + plist[ tetralist[tno].p4 ].z ) * 0.25;
}

bool ModuleData::tetraeder_neigh( int tno1, int tno2 ) const
{
   POINT cp1, cp2;
   double r1, r2, R;

   tetraeder_center( tno1, cp1 );
   tetraeder_center( tno2, cp2 );

   r1 = HYPOT( cp1.x-plist[ tetralist[tno1].p1 ].x,
               cp1.y-plist[ tetralist[tno1].p1 ].y,
               cp1.z-plist[ tetralist[tno1].p1 ].z  );
   R = HYPOT( cp1.x-plist[ tetralist[tno1].p2 ].x,
              cp1.y-plist[ tetralist[tno1].p2 ].y,
              cp1.z-plist[ tetralist[tno1].p2 ].z  );
   r1 = ( R > r1 ? R : r1 );
   R = HYPOT( cp1.x-plist[ tetralist[tno1].p3 ].x,
              cp1.y-plist[ tetralist[tno1].p3 ].y,
              cp1.z-plist[ tetralist[tno1].p3 ].z  );
   r1 = ( R > r1 ? R : r1 );
   R = HYPOT( cp1.x-plist[ tetralist[tno1].p4 ].x,
              cp1.y-plist[ tetralist[tno1].p4 ].y,
              cp1.z-plist[ tetralist[tno1].p4 ].z  );
   r1 = ( R > r1 ? R : r1 );

   r2 = HYPOT( cp1.x-plist[ tetralist[tno2].p1 ].x,
               cp1.y-plist[ tetralist[tno2].p1 ].y,
               cp1.z-plist[ tetralist[tno2].p1 ].z  );
   R = HYPOT( cp1.x-plist[ tetralist[tno2].p2 ].x,
              cp1.y-plist[ tetralist[tno2].p2 ].y,
              cp1.z-plist[ tetralist[tno2].p2 ].z  );
   r2 = ( R > r2 ? R : r2 );
   R = HYPOT( cp1.x-plist[ tetralist[tno2].p3 ].x,
              cp1.y-plist[ tetralist[tno2].p3 ].y,
              cp1.z-plist[ tetralist[tno2].p3 ].z  );
   r2 = ( R > r2 ? R : r2 );
   R = HYPOT( cp1.x-plist[ tetralist[tno2].p4 ].x,
              cp1.y-plist[ tetralist[tno2].p4 ].y,
              cp1.z-plist[ tetralist[tno2].p4 ].z  );
   r2 = ( R > r2 ? R : r2 );

   if ( HYPOT( cp1.x-cp2.x, cp1.y-cp2.y, cp1.z-cp2.z ) > r1+r2 ) return(false);

   return(true);

}

double ModuleData::tetraeder_volume( int tno ) const
{
   double a[3], b[3], c[3], nx, ny, nz;

#ifdef BOUNDARY_CHECK
   if ( tno < 0 || tno >= (int)tetralist.size() )
   {
      cerr << "Error in ModuleData::tetraeder_volume(): tno out of bounds!"
           << endl;
      exit(1);
   }
#endif

   a[0] = plist[ tetralist[tno].p2 ].x - plist[ tetralist[tno].p1 ].x;
   a[1] = plist[ tetralist[tno].p2 ].y - plist[ tetralist[tno].p1 ].y;
   a[2] = plist[ tetralist[tno].p2 ].z - plist[ tetralist[tno].p1 ].z;

   b[0] = plist[ tetralist[tno].p3 ].x - plist[ tetralist[tno].p1 ].x;
   b[1] = plist[ tetralist[tno].p3 ].y - plist[ tetralist[tno].p1 ].y;
   b[2] = plist[ tetralist[tno].p3 ].z - plist[ tetralist[tno].p1 ].z;

   c[0] = plist[ tetralist[tno].p4 ].x - plist[ tetralist[tno].p1 ].x;
   c[1] = plist[ tetralist[tno].p4 ].y - plist[ tetralist[tno].p1 ].y;
   c[2] = plist[ tetralist[tno].p4 ].z - plist[ tetralist[tno].p1 ].z;

   nx = a[1]*b[2] - a[2]*b[1];
   ny = a[2]*b[0] - a[0]*b[2];
   nz = a[0]*b[1] - a[1]*b[0];

   return( fabs(nx*c[0] + ny*c[1] + nz*c[2])/6.0 );
}

int ModuleData::in_tetraeder( POINT &p ) const
{
   int t;
   double V6, a[3], b[3], c[3], nx, ny, nz;

   for ( t=0; t<(int)tetralist.size(); ++t )
   {
      V6 = 0.0;

// interface 1
      a[0] = plist[ tetralist[t].p1 ].x - p.x;
      a[1] = plist[ tetralist[t].p1 ].y - p.y;
      a[2] = plist[ tetralist[t].p1 ].z - p.z;

      b[0] = plist[ tetralist[t].p2 ].x - p.x;
      b[1] = plist[ tetralist[t].p2 ].y - p.y;
      b[2] = plist[ tetralist[t].p2 ].z - p.z;

      nx = a[1]*b[2] - a[2]*b[1];
      ny = a[2]*b[0] - a[0]*b[2];
      nz = a[0]*b[1] - a[1]*b[0];

// subtetraeder 1
      c[0] = plist[ tetralist[t].p3 ].x - p.x;
      c[1] = plist[ tetralist[t].p3 ].y - p.y;
      c[2] = plist[ tetralist[t].p3 ].z - p.z;

      V6 += fabs(nx*c[0] + ny*c[1] + nz*c[2]);

// subtetraeder 2
      c[0] = plist[ tetralist[t].p4 ].x - p.x;
      c[1] = plist[ tetralist[t].p4 ].y - p.y;
      c[2] = plist[ tetralist[t].p4 ].z - p.z;

      V6 += fabs(nx*c[0] + ny*c[1] + nz*c[2]);

// interface 2
      a[0] = plist[ tetralist[t].p3 ].x - p.x;
      a[1] = plist[ tetralist[t].p3 ].y - p.y;
      a[2] = plist[ tetralist[t].p3 ].z - p.z;

      b[0] = plist[ tetralist[t].p4 ].x - p.x;
      b[1] = plist[ tetralist[t].p4 ].y - p.y;
      b[2] = plist[ tetralist[t].p4 ].z - p.z;

      nx = a[1]*b[2] - a[2]*b[1];
      ny = a[2]*b[0] - a[0]*b[2];
      nz = a[0]*b[1] - a[1]*b[0];

// subtetraeder 1
      c[0] = plist[ tetralist[t].p1 ].x - p.x;
      c[1] = plist[ tetralist[t].p1 ].y - p.y;
      c[2] = plist[ tetralist[t].p1 ].z - p.z;

      V6 += fabs(nx*c[0] + ny*c[1] + nz*c[2]);

// subtetraeder 2
      c[0] = plist[ tetralist[t].p2 ].x - p.x;
      c[1] = plist[ tetralist[t].p2 ].y - p.y;
      c[2] = plist[ tetralist[t].p2 ].z - p.z;

      V6 += fabs(nx*c[0] + ny*c[1] + nz*c[2]);

      if ( fabs( V6 - 6.0*tetraeder_volume(t) ) < V6*10.0*DBL_EPSILON ) return(t);
   }

   return(-1);
}
