/*
*  Integral of singularity on cylinder
*  (c) 26.2.2007 Ingo D. Rullhusen
*/

#include <math.h>
#include <float.h>


static double SQU( double x ){ return(x*x); }


double Psing( int n, double *bx, double *by,
              double rx, double ry, double rz )
{
   int i;
   double sg, nv, sig, p0i, ui[2], li[2], rp[3], rm[3], r0[3],
          rpi, rmi, r0iq, lpi, lmi, logt, d;

   sg = 0.0;

   for ( i=0; i<n-1; i++ )
   {

      li[0] = bx[i+1] - bx[i];
      li[1] = by[i+1] - by[i];
      nv = hypot( li[0], li[1] );
      li[0] /= nv;
      li[1] /= nv;

      ui[0] =  li[1];
      ui[1] = -li[0];

      rp[0] = bx[i+1] - rx;
      rp[1] = by[i+1] - ry;
      rp[2] =         - rz;

      rpi = sqrt( SQU(rp[0]) + SQU(rp[1]) + SQU(rp[2]) );
      lpi = rp[0]*li[0] + rp[1]*li[1];

      rm[0] = bx[i] - rx;
      rm[1] = by[i] - ry;
      rm[2] =       - rz;

      rmi = sqrt( SQU(rm[0]) + SQU(rm[1]) + SQU(rm[2]) );
      lmi = rm[0]*li[0] + rm[1]*li[1];

      nv = rm[0]*li[0] + rm[1]*li[1];
      r0[0] = rm[0] - nv * li[0];
      r0[1] = rm[1] - nv * li[1];
      r0[2] = rm[2];

      r0iq = SQU(r0[0]) + SQU(r0[1]) + SQU(r0[2]);

      p0i = ui[0]*rm[0] + ui[1]*rm[1];
      sig = ( p0i < 0.0 ? -1.0 : +1.0 );
      p0i = fabs( p0i );

      d = fabs(rz);

      logt = 0.0;
      if ( fabs(rpi+lpi) > DBL_EPSILON ) logt += p0i * log( rpi + lpi );
      if ( fabs(rmi+lmi) > DBL_EPSILON ) logt -= p0i * log( rmi + lmi );

      sg += sig * ( logt - d * ( atan2( p0i*lpi, r0iq+d*rpi )
                                 - atan2( p0i*lmi, r0iq+d*rmi ) ) );

   }

   return(sg);
}


double IntInvR( double tx[4], double ty[4], double tz[4], 
                double apx, double apy, double apz )
{
   const int pno = 4;
   double px[pno], py[pno];
   double x[3], y[3], z[3], lx, ly_x, ly_y, lz_x, lz_y, lz_z, d_j, I;

   int f;
   const int fpidx1[4] = { 1, 0, 0, 0 };
   const int fpidx2[4] = { 2, 2, 1, 1 };
   const int fpidx3[4] = { 3, 3, 3, 2 };

   I = 0.0;

   for ( f=0; f<4; f++ )
   {
      x[0] = tx[ fpidx2[f] ] - tx[ fpidx1[f] ];
      x[1] = ty[ fpidx2[f] ] - ty[ fpidx1[f] ];
      x[2] = tz[ fpidx2[f] ] - tz[ fpidx1[f] ];
      lx = sqrt( x[0]*x[0] + x[1]*x[1] + x[2]*x[2] );
      x[0] /= lx;
      x[1] /= lx;
      x[2] /= lx;

      y[0] = tx[ fpidx3[f] ] - tx[ fpidx1[f] ];
      y[1] = ty[ fpidx3[f] ] - ty[ fpidx1[f] ];
      y[2] = tz[ fpidx3[f] ] - tz[ fpidx1[f] ];
      ly_x = x[0]*y[0] + x[1]*y[1] + x[2]*y[2];
      y[0] -= ly_x*x[0];
      y[1] -= ly_x*x[1];
      y[2] -= ly_x*x[2];
      ly_y = sqrt( y[0]*y[0] + y[1]*y[1] + y[2]*y[2] );
      y[0] /= ly_y;
      y[1] /= ly_y;
      y[2] /= ly_y;

      z[0] = tx[ fpidx1[f] ] - tx[f];
      z[1] = ty[ fpidx1[f] ] - ty[f];
      z[2] = tz[ fpidx1[f] ] - tz[f];
      lz_x = x[0]*z[0] + x[1]*z[1] + x[2]*z[2];
      z[0] -= lz_x*x[0];
      z[1] -= lz_x*x[1];
      z[2] -= lz_x*x[2];
      lz_y = y[0]*z[0] + y[1]*z[1] + y[2]*z[2];
      z[0] -= lz_y*y[0];
      z[1] -= lz_y*y[1];
      z[2] -= lz_y*y[2];
      lz_z = sqrt( z[0]*z[0] + z[1]*z[1] + z[2]*z[2] );
      z[0] /= lz_z;
      z[1] /= lz_z;
      z[2] /= lz_z;

      d_j = z[0]*(apx-tx[ fpidx1[f] ]) + z[1]*(apy-ty[ fpidx1[f] ]) + z[2]*(apz-tz[ fpidx1[f] ]);

      px[0] = 0.0;   py[0] = 0.0;
      px[1] = lx;    py[1] = 0.0;
      px[2] = ly_x;  py[2] = ly_y;
      px[3] = px[0]; py[3] = py[0];

      I += d_j * Psing(
              pno, px, py, 
              x[0]*(apx-tx[ fpidx1[f] ]) + x[1]*(apy-ty[ fpidx1[f] ]) + x[2]*(apz-tz[ fpidx1[f] ]),
              y[0]*(apx-tx[ fpidx1[f] ]) + y[1]*(apy-ty[ fpidx1[f] ]) + y[2]*(apz-tz[ fpidx1[f] ]),
              fabs(d_j)  // not RH sense possible, doesn't matter
                      );
   }

   return( -0.5*I );
}
